program DumbTTY;
{ A small dumb terminal emulation  Copyright (c) 1997, Marius Gedminas }
{ This code is free for non-commercial use. But you must use it at your}
{ own risk: I will not be liable for anything concerning it.           }
{                                                                      }
{ Sorry for the lack of comments...                                    }

uses
  ComIO, Video;

  { - Keyboard - }

const
  LastKey: Word = 0;

function GetKey: Word; assembler;
asm
  mov  ah, 0
  int  16h
  mov  LastKey, ax
end;

function IsKey: Boolean; assembler;
asm
  mov   ah, 1
  int   16h
  mov   ax, False
  jz    @@1
  mov   ax, True
@@1:
end;

function PeekKey: Word; assembler;
asm
  mov   ah, 1
  int   16h
  jnz   @@1
  xor   ax, ax
@@1:
end;

function InputStr(x, y, w: Integer; var S: string; MaxLen: Integer): Boolean;
var
  B: TDrawBuffer;
  First: Boolean;
  Pos: Integer;
  OldCur: Word;
  OldX, OldY: Byte;
  InsMode: Boolean;

  procedure Draw(Attr: Byte);
  begin
    MoveChar(B, ' ', $0F, w);
    MoveStr(B[1], S, Attr);
    WriteLine(x, y, w, 1, B);
    SetCursorPos(x+Pos+1, y);
  end;

begin
  OldCur := GetCursor; GetCursorPos(OldX, OldY);
  Pos := Length(S); InsMode := True;
  Draw($1F);
  SetCursor(StdCursor);

  First := True;
  repeat
    case GetKey of
      $1C0D: { Enter } Break;
      $011B: { Esc   } Break;
      $0E08: { Back  }
        begin
          First := False;
          if Pos > 0 then
            begin Delete(S, Pos, 1); Dec(Pos);
            end;
          Draw($0F);
        end;
      $5300: { Del }
        begin
          First := False;
          Delete(S, Pos+1, 1);
          Draw($0F);
        end;
      $5200: { Ins }
        begin
          First := False;
          InsMode := not InsMode;
          Draw($0F);
          if InsMode then SetCursor(StdCursor) else SetCursor(BigCursor);
        end;
      $4700: { Home }
        begin
          First := False;
          Pos := 0;
          Draw($0F);
        end;
      $4F00: { End }
        begin
          First := False;
          Pos := Length(S);
          Draw($0F);
        end;
      $4B00: { Left }
        begin
          First := False;
          if Pos > 0 then Dec(Pos);
          Draw($0F);
        end;
      $4D00: { Right }
        begin
          First := False;
          if Pos < Length(S) then Inc(Pos);
          Draw($0F);
        end;
      $4800, $4900, $4C00, $5000, $5100: { Other cursor keys }
        if First then
          begin
            First := False;
            Draw($0F);
          end;
      else
        if Lo(LastKey) <> 0 then
          begin
            if First then begin S := ''; Pos := 0; First := False; end;
            if not InsMode and (Pos < Length(S)) then
              begin
                Inc(Pos); S[Pos] := Char(LastKey);
              end
            else
            if Length(S) < MaxLen then
              begin
                Inc(Pos); Insert(Char(LastKey), S, Pos);
              end;
            Draw($0F);
          end;
    end;
  until False;
  InputStr := (LastKey = $1C0D);
  SetCursor(OldCur); SetCursorPos(OldX, OldY);
end;

  { - Timer - }

procedure Wait(Ticks: Word);
var
  Time: Longint;
begin
  Time := GetTimer;
  repeat
    if GetTimer < Time then Time := GetTimer; { Midnight }
  until GetTimer >= Time+Ticks;
end;

  { - Terminal - }

type
  TTermWin = object
    Origin, Size: TPoint;
    BufSize: Word;
    Buf: PScrBuf;
    BufPos: Word;
    Cursor: TPoint;
    constructor Init(x, y, w, h: Integer);
    destructor Done;
    procedure Draw;
    procedure Write(B: Byte);
  end;

constructor TTermWin.Init(x, y, w, h: Integer);
begin
  Origin.x := x; Origin.y := y;
  Size.x := w; Size.y := h;
  BufSize := w*h;
  GetMem(Buf, 2*BufSize);
  MoveChar(Buf^, ' ', $07, BufSize);
  Cursor.x := 0; Cursor.y := 0;
  BufPos := 0;
end;

destructor TTermWin.Done;
begin
  FreeMem(Buf, 2*BufSize);
end;

procedure TTermWin.Draw;
begin
  WriteBuf(Origin.x, Origin.y, Size.x, Size.Y, Buf^);
  SetCursorPos(Origin.x+Cursor.x, Origin.y+Cursor.y);
end;

procedure TTermWin.Write(B: Byte);

  procedure ScrollUp;
  begin
    Move(Buf^[Size.X], Buf^[0], (BufSize-Size.X)*2);
    MoveChar(Buf^[BufSize-Size.X], ' ', $07, Size.X);
    Dec(Cursor.Y); Dec(BufPos, Size.X);
  end;

begin
  case B of
    $0D: { CR }
      begin
        Cursor.x := 0;
        BufPos := Cursor.y*Size.x;
      end;
    $0A: { LF }
      begin
        Inc(Cursor.y); Inc(BufPos, Size.x);
        if Cursor.y >= Size.y then ScrollUp;
      end;
    { Check others: BEL, BS, TAB... }
    else
      begin
        WordRec(Buf^[BufPos]).Lo := B;
        Inc(Cursor.x); Inc(BufPos);
        if Cursor.x >= Size.x then
          begin
            Cursor.x := 0; Inc(Cursor.y);
            BufPos := Cursor.y*Size.x;
            if Cursor.y >= Size.y then ScrollUp;
          end;
      end;
  end;
  Draw;
end;

  { - Modem - }

const
  Parity: array[TParity] of Char = 'noe';

const
  { - Line settings - }
  ComPort: Byte = 2; { COM2 }
  ComSpeed: Longint = 9600; { Baud rate }
  ComData: Byte = 8;
  ComParity: TParity = ParityNone;
  ComStop: Byte = 1; { 8n1 connection }
  { - Modem settings - }
  InitStr: string = ^@'ATZ'^M;
  DialStr: string = 'ATDP'; { ATDP for pulse dialing, 'ATDT' for tone dialing }
  DialSuffix: string = ^M;
  EscapeStr: string = ^@'+++'^@;
  HangupStr: string = 'ATH0'^M;
  { - User settings - }
  PhoneNumber: string = '';
  { - Terminal settings - }
  LocalEcho: Boolean = False;

function ErrorStr(Code: Integer): string;
begin
  case Code of
    errOk: ErrorStr := 'Ok';
    errBadSettings: ErrorStr := 'Bad line settings';
    errCantSend: ErrorStr := 'Cannot send';
    errTimeout: ErrorStr := 'Timeout';
    errOverflow: ErrorStr := 'Buffer overflow';
    else ErrorStr := 'Unknown error';
  end;
end;

var
  C: TComPort;
  T: TTermWin;
  B: Byte;
  Status: Byte;
  Sending: Boolean;

function AbortIO: Boolean; far;
begin
  AbortIO := False;
  if PeekKey <> $011B { Esc } then Exit;
  GetKey;
  AbortIO := True;
end;

procedure DrawInfoLine;
const
  Copyr = 'DumbTTY v1.00 ';
var
  B: TDrawBuffer;
  S: string[7];
  Cl: Byte;
begin
  Status := C.ModemStatus;
  {0....x....1....x....2....x....3....x....4}
  { COMx 115200, 8n1  CD RI DSR CTS  Error}
  MoveChar(B, ' ', $30, ScrW);

  MoveStr(B[1], 'COM', $30); WordRec(B[4]).Lo := Ord('0')+ComPort;
  Str(ComSpeed:6, S); MoveStr(B[6], S, $30);
  WordRec(B[12]).Lo := Ord(',');
  WordRec(B[14]).Lo := Ord('0')+ComData;
  WordRec(B[15]).Lo := Ord(Parity[ComParity]);
  WordRec(B[16]).Lo := Ord('0')+ComStop;
  if Status and msDCD <> 0 then Cl := $3F else Cl := $38; MoveStr(B[19], 'CD', Cl);
  if Status and msRI  <> 0 then Cl := $3F else Cl := $38; MoveStr(B[22], 'RI', Cl);
  if Status and msDSR <> 0 then Cl := $3F else Cl := $38; MoveStr(B[25], 'DSR', Cl);
  if Status and msCTS <> 0 then Cl := $3F else Cl := $38; MoveStr(B[29], 'CTS', Cl);
  if C.Error <> 0 then
    MoveStr(B[34], ErrorStr(C.Error), $3E);
  if Sending then WordRec(B[0]).Lo := Ord('>');

  MoveStr(B[ScrW-Length(Copyr)], Copyr, $3F);
  WriteLine(0, ScrH-1, ScrW, 1, B);
end;

procedure SendStr(const S: string);
var
  I: Byte;
begin
  C.Error := errOk;
  for I := 1 to Length(S) do
    if S[I] = ^@ then
      Wait(18)
    else
      begin
        if LocalEcho then T.Write(Ord(S[I]));
        Sending := True; DrawInfoLine;
        C.Send(Ord(S[I]));
        Sending := False; DrawInfoLine;
      end;
end;

procedure Help;
const
  w = 54;
  h = 17+2;
  Title = ' Help ';
  Line: array[1..h-2] of string[w+10] = (
    '',
    'A small dumb terminal demo by ~Marius Gedminas~',
    '',
    '~F1~               Help',
    '',
    '~Alt-L~            Line settings',
    '~Alt-M~            Modem settings',
    '~Alt-T~            Terminal settings',
    '~Alt-V~            Local echo',
    '',
    '~Alt-I~            Init',
    '~Alt-D~            Dial',
    '~Alt-E~            Escape to command mode',
    '~Alt-H~            Hangup',
    '',
    '~Alt-X~            Quit',
    '');

var
  x, y, I: Integer;
  B: TDrawBuffer;
  Buf: Pointer;
begin
  x := (ScrW-w) div 2; y := (ScrH-h) div 2;
  GetMem(Buf, (w+Aspect)*(h+1)*2); SaveBuf(x, y, w+Aspect, h+1, Buf^);
  FillAttr(x+w, y+1, Aspect, h-1, $08);
  FillAttr(x+Aspect, y+h, w, 1, $08);

  MoveChar(B[0], '', $2F, 1);
  MoveChar(B[1], '', $2F, w-2);
  MoveChar(B[w-1], '', $2F, 1);
  MoveStr(B[(w-Length(Title)) div 2], Title, $2E);
  WriteLine(x, y, w, 1, B);

  for I := 1 to h-2 do
    begin
      MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
      if Line[I][1] = ^C then
        MoveCStr(B[(w-CStrLen(Line[I])+1) div 2], Copy(Line[I], 2, 255), $2E2F)
      else
        MoveCStr(B[3], Line[I], $2E2F);
      WriteLine(x, y+I, w, 1, B);
    end;

  MoveChar(B[0], '', $2F, 1);
  MoveChar(B[1], '', $2F, w-2);
  MoveChar(B[w-1], '', $2F, 1);
  WriteLine(x, y+h-1, w, 1, B);

  GetKey;
  WriteBuf(x, y, w+Aspect, h+1, Buf^); FreeMem(Buf, (w+Aspect)*(h+1)*2);
end;

procedure LineSettings;
const
  w = 60;
  h = 9;
  Title = ' Line settings ';
var
  x, y: Integer;
  B: TDrawBuffer;
  Buf: Pointer;
  OldPort, OldData, OldStop: Byte; OldSpeed: Longint; OldParity: TParity;
  S: string; Temp: Longint; I: Integer;

  procedure Draw;
  begin
    MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
    MoveCStr(B[3], '~C~OM port:    ~( ) ~COM1~  ( ) ~COM2~  ( ) ~COM3~  ( ) ~COM4~', $2E2F);
    WordRec(B[17+(ComPort-1)*10]).Lo := 7;
    WriteLine(x, y+2, w, 1, B);

    MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
    MoveCStr(B[3], '~S~peed:', $2E2F);
    Str(ComSpeed, S); MoveStr(B[16], S, $2F);
    WriteLine(x, y+3, w, 1, B);

    MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
    MoveCStr(B[3], '~D~ata bits:   ~( ) ~5~     ( ) ~6~     ( ) ~7~     ( ) ~8~', $2E2F);
    WordRec(B[17+(ComData-5)*10]).Lo := 7;
    WriteLine(x, y+4, w, 1, B);

    MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
    MoveCStr(B[3], '~P~arity:      ~( ) ~None~  ( ) ~Odd~   ( ) ~Even~', $2E2F);
    WordRec(B[17+Ord(ComParity)*10]).Lo := 7;
    WriteLine(x, y+5, w, 1, B);

    MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
    MoveCStr(B[3], 'S~t~op bit(s): ~( ) ~1~     ( ) ~2~', $2E2F);
    WordRec(B[17+(ComStop-1)*10]).Lo := 7;
    WriteLine(x, y+6, w, 1, B);
  end;

begin
  OldPort := ComPort; OldData := ComData; OldStop := ComStop;
  OldSpeed := ComSpeed; OldParity := ComParity;

  x := (ScrW-w) div 2; y := (ScrH-h) div 2;
  GetMem(Buf, (w+Aspect)*(h+1)*2); SaveBuf(x, y, w+Aspect, h+1, Buf^);
  FillAttr(x+w, y+1, Aspect, h-1, $08);
  FillAttr(x+Aspect, y+h, w, 1, $08);

  MoveChar(B[0], '', $2F, 1);
  MoveChar(B[1], '', $2F, w-2);
  MoveChar(B[w-1], '', $2F, 1);
  MoveStr(B[(w-Length(Title)) div 2], Title, $2E);
  WriteLine(x, y, w, 1, B);

  MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
  WriteLine(x, y+1, w, 1, B);

  Draw;

  MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
  WriteLine(x, y+7, w, h-7, B);

  MoveChar(B[0], '', $2F, 1);
  MoveChar(B[1], '', $2F, w-2);
  MoveChar(B[w-1], '', $2F, 1);
  WriteLine(x, y+h-1, w, 1, B);

  repeat
    case GetKey of
      $011B: { Esc }
        begin
        { Restore? No, I think most people will exit these dialogs with Esc
          ComPort := OldPort; ComData := OldData; ComStop := OldStop;
          ComSpeed := OldSpeed; ComParity := OldParity;
          Draw; }
          Break;
        end;
      $1C0D: { Enter }
        Break;
      else
        case UpCase(Char(LastKey)) of
          'C': { Port }
            begin
              if ComPort = 4 then ComPort := 1 else Inc(ComPort);
              Draw;
            end;
          'S': { Speed }
            begin
              Str(ComSpeed, S);
              if InputStr(x+16, y+3, w-19, S, 15) then
                begin
                  Val(S, Temp, I);
                  if I = 0 then ComSpeed := Temp;
                  Draw;
                end;
            end;
          'D': { Data }
            begin
              if ComData = 8 then ComData := 5 else Inc(ComData);
              Draw;
            end;
          'P': { Parity }
            begin
              if ComParity = High(TParity) then ComParity := Low(TParity) else Inc(ComParity);
              Draw;
            end;
          'T': { Stop }
            begin
              if ComStop = 2 then ComStop := 1 else Inc(ComStop);
              Draw;
            end;
        end;
    end;
  until False;

  WriteBuf(x, y, w+Aspect, h+1, Buf^); FreeMem(Buf, (w+Aspect)*(h+1)*2);

  if ComPort <> OldPort then
    begin
      C.Done;
      C.Init(ComPort, comInt, ComSpeed, ComData, ComParity, ComStop);
      @C.AbortIO := @AbortIO;
    end
  else
  if (ComSpeed <> OldSpeed) or (ComData <> OldData) or
     (ComParity <> OldParity) or (ComStop <> OldStop) then
    C.Setup(ComSpeed, ComData, ComParity, ComStop);
end;

procedure ModemSettings;
const
  w = 60;
  h = 11;
  Title = ' Modem settings ';
var
  x, y: Integer;
  B: TDrawBuffer;
  Buf: Pointer;
  OldInit, OldDial, OldDSuff, OldEsc, OldHang: string;
  Temp: string;

  procedure PreEdit(var S: string);
  var
    I: Integer;
  begin
    I := 1;
    while I <= Length(S) do
      begin
        if S[I] = ^@ then begin Insert('\', S, I); Inc(I); S[I] := 'w' end
        else
        if S[I] = ^M then begin Insert('\', S, I); Inc(I); S[I] := 'r' end
        else
        if S[I] = '\' then begin Insert('\', S, I); Inc(I); end;
        Inc(I);
      end;
  end;
  procedure PostEdit(var S: string);
  var
    I: Integer;
  begin
    I := 1;
    while I < Length(S) do
      begin
        if S[I] = '\' then
          begin
            Delete(S, I, 1);
            case S[I] of
              'w': S[I] := ^@;
              'r': S[I] := ^M;
            end;
          end;
        Inc(I);
      end;
  end;

  procedure Draw;
  begin
    MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
    MoveCStr(B[3], '~I~nit:', $2E2F);
    MoveStr(B[16], InitStr, $2F);
    WriteLine(x, y+2, w, 1, B);

    MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
    MoveCStr(B[3], '~D~ial prefix:', $2E2F);
    MoveStr(B[16], DialStr, $2F);
    WriteLine(x, y+3, w, 1, B);

    MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
    MoveCStr(B[3], 'Dial ~s~uffix:', $2E2F);
    MoveStr(B[16], DialSuffix, $2F);
    WriteLine(x, y+4, w, 1, B);

    MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
    MoveCStr(B[3], '~E~scape:', $2E2F);
    MoveStr(B[16], EscapeStr, $2F);
    WriteLine(x, y+5, w, 1, B);

    MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
    MoveCStr(B[3], '~H~angup:', $2E2F);
    MoveStr(B[16], HangupStr, $2F);
    WriteLine(x, y+6, w, 1, B);
  end;

begin
  PreEdit(InitStr); PreEdit(DialStr); PreEdit(DialSuffix);
  PreEdit(EscapeStr); PreEdit(HangupStr);

  OldInit := InitStr; OldDial := DialStr; OldDSuff := DialSuffix;
  OldEsc := EscapeStr; OldHang := HangupStr;

  x := (ScrW-w) div 2; y := (ScrH-h) div 2;
  GetMem(Buf, (w+Aspect)*(h+1)*2); SaveBuf(x, y, w+Aspect, h+1, Buf^);
  FillAttr(x+w, y+1, Aspect, h-1, $08);
  FillAttr(x+Aspect, y+h, w, 1, $08);

  MoveChar(B[0], '', $2F, 1);
  MoveChar(B[1], '', $2F, w-2);
  MoveChar(B[w-1], '', $2F, 1);
  MoveStr(B[(w-Length(Title)) div 2], Title, $2E);
  WriteLine(x, y, w, 1, B);

  MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
  WriteLine(x, y+1, w, 1, B);

  Draw;

  MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
  WriteLine(x, y+7, w, 1, B);

  MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
  MoveStr(B[3], '\w - wait 1 second  \r - send a CR', $2A);
  WriteLine(x, y+8, w, 1, B);

  MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
  WriteLine(x, y+9, w, h-9, B);

  MoveChar(B[0], '', $2F, 1);
  MoveChar(B[1], '', $2F, w-2);
  MoveChar(B[w-1], '', $2F, 1);
  WriteLine(x, y+h-1, w, 1, B);

  repeat
    case GetKey of
      $011B: { Esc }
        begin
        { Restore? No, I think most people will exit these dialogs with Esc
          InitStr := OldInit; DialStr := OldDial; DialSuffix := OldDSuff;
          EscapeStr := OldEsc; HangupStr := OldHang;
          Draw; }
          Break;
        end;
      $1C0D: { Enter }
        Break;
      else
        case UpCase(Char(LastKey)) of
          'I': { Init }
            begin
              Temp := InitStr;
              if InputStr(x+16, y+2, w-19, Temp, 40) then
                InitStr := Temp;
              Draw;
            end;
          'D': { Dial }
            begin
              Temp := DialStr;
              if InputStr(x+16, y+3, w-19, Temp, 40) then
                DialStr := Temp;
              Draw;
            end;
          'S': { Dial suffix }
            begin
              Temp := DialSuffix;
              if InputStr(x+16, y+4, w-19, Temp, 40) then
                DialSuffix := Temp;
              Draw;
            end;
          'E': { Escape }
            begin
              Temp := EscapeStr;
              if InputStr(x+16, y+5, w-19, Temp, 40) then
                EscapeStr := Temp;
              Draw;
            end;
          'H': { Hangup }
            begin
              Temp := HangupStr;
              if InputStr(x+16, y+6, w-19, Temp, 40) then
                HangupStr := Temp;
              Draw;
            end;
        end;
    end;
  until False;

  WriteBuf(x, y, w+Aspect, h+1, Buf^); FreeMem(Buf, (w+Aspect)*(h+1)*2);

  PostEdit(InitStr); PostEdit(DialStr); PostEdit(DialSuffix);
  PostEdit(EscapeStr); PostEdit(HangupStr);
end;

procedure TermSettings;
const
  w = 60;
  h = 5;
  Title = ' Terminal settings ';
var
  x, y: Integer;
  B: TDrawBuffer;
  Buf: Pointer;
  OldLEcho: Boolean;

  procedure Draw;
  begin
    MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
    MoveCStr(B[3], '~L~ocal echo:  ~( ) ~Off~   ( ) ~On~', $2E2F);
    WordRec(B[17+Ord(LocalEcho)*10]).Lo := 7;
    WriteLine(x, y+2, w, 1, B);
  end;

begin
  OldLEcho := LocalEcho;

  x := (ScrW-w) div 2; y := (ScrH-h) div 2;
  GetMem(Buf, (w+Aspect)*(h+1)*2); SaveBuf(x, y, w+Aspect, h+1, Buf^);
  FillAttr(x+w, y+1, Aspect, h-1, $08);
  FillAttr(x+Aspect, y+h, w, 1, $08);

  MoveChar(B[0], '', $2F, 1);
  MoveChar(B[1], '', $2F, w-2);
  MoveChar(B[w-1], '', $2F, 1);
  MoveStr(B[(w-Length(Title)) div 2], Title, $2E);
  WriteLine(x, y, w, 1, B);

  MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
  WriteLine(x, y+1, w, 1, B);

  Draw;

  MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
  WriteLine(x, y+3, w, h-3, B);

  MoveChar(B[0], '', $2F, 1);
  MoveChar(B[1], '', $2F, w-2);
  MoveChar(B[w-1], '', $2F, 1);
  WriteLine(x, y+h-1, w, 1, B);

  repeat
    case GetKey of
      $011B: { Esc }
        begin
        { Restore? No, I think most people will exit these dialogs with Esc
          LocalEcho := OldLEcho;
          Draw; }
          Break;
        end;
      $1C0D: { Enter }
        Break;
      else
        case UpCase(Char(LastKey)) of
          'L': { Local echo }
            begin
              LocalEcho := not LocalEcho;
              Draw;
            end;
        end;
    end;
  until False;

  WriteBuf(x, y, w+Aspect, h+1, Buf^); FreeMem(Buf, (w+Aspect)*(h+1)*2);
end;

function GetPhone(var Phn: string; MaxLen: Byte): Boolean;
const
  w = 60;
  h = 6;
  Title = ' Dial ';
  Text = 'Enter phone number';
var
  x, y: Integer;
  B: TDrawBuffer;
  Buf: Pointer;
begin
  x := (ScrW-w) div 2; y := (ScrH-h) div 2;
  GetMem(Buf, (w+Aspect)*(h+1)*2); SaveBuf(x, y, w+Aspect, h+1, Buf^);
  FillAttr(x+w, y+1, Aspect, h-1, $08);
  FillAttr(x+Aspect, y+h, w, 1, $08);

  MoveChar(B[0], '', $2F, 1);
  MoveChar(B[1], '', $2F, w-2);
  MoveChar(B[w-1], '', $2F, 1);
  MoveStr(B[(w-Length(Title)) div 2], Title, $2E);
  WriteLine(x, y, w, 1, B);

  MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
  WriteLine(x, y+1, w, 1, B);

  MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
  MoveStr(B[3], Text, $2F);
  WriteLine(x, y+2, w, 1, B);

  MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
  MoveChar(B[3], ' ', $0F, w-6);
  WriteLine(x, y+3, w, 1, B);

  MoveChar(B[0], '', $2F, 1); MoveChar(B[1], ' ', $2F, w-2); MoveChar(B[w-1], '', $2F, 1);
  WriteLine(x, y+4, w, h-5, B);

  MoveChar(B[0], '', $2F, 1);
  MoveChar(B[1], '', $2F, w-2);
  MoveChar(B[w-1], '', $2F, 1);
  WriteLine(x, y+h-1, w, 1, B);

  GetPhone := InputStr(x+3, y+3, w-6, Phn, MaxLen);

  WriteBuf(x, y, w+Aspect, h+1, Buf^); FreeMem(Buf, (w+Aspect)*(h+1)*2);
end;

var
  OldExit: Pointer;
  OldCursor: Word;

procedure Finalize; far;
begin
  ExitProc := OldExit;
  ClearScreen;
  SetCursorPos(0, 0);
  SetCursor(OldCursor);
end;

begin
  OldCursor := GetCursor; OldExit := ExitProc; ExitProc := @Finalize;
  ClearScreen; Status := 0;

  C.Init(ComPort, comInt, ComSpeed, ComData, ComParity, ComStop);
  @C.AbortIO := @AbortIO;

  T.Init(0, 0, ScrW, ScrH-1); T.Draw;
  DrawInfoLine;
  repeat
    DrawInfoLine;
    if C.Pending then
      begin
        C.Receive(B);
        T.Write(B);
      end;
    if IsKey then
      case GetKey of
        $3B00: { F1    } Help;

        $2600: { Alt-L } LineSettings;
        $3200: { Alt-M } ModemSettings;
        $1400: { Alt-T } TermSettings;
        $2F00: { Alt-V } begin LocalEcho := not LocalEcho; end;

        $1700: { Alt-I } SendStr(InitStr);
        $2000: { Alt-D }
          if GetPhone(PhoneNumber, 40) then
            SendStr(DialStr+PhoneNumber+DialSuffix);
        $1200: { Alt-E } SendStr(EscapeStr);
        $2300: { Alt-H } SendStr(HangupStr);

        $2D00: { Alt-X } Break;
        else
          if Lo(LastKey) <> 0 then
            SendStr(Char(LastKey));
      end;
  until False;

  C.Done;
end.